/*-
 * Copyright (C)2005..2024 @BABOLO http://www.babolo.ru/
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ident "@(#)Copyright (C)2005..2024 @BABOLO http://www.babolo.ru/"
#ident "$Id: insert.c,v 1.15 2024/04/14 20:41:05 babolo Exp $"

#define BLIN_COMPAT  4
#define MULAR_COMPAT VMAJOR
#define MULAR_INTERNAL

#include <sys/types.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <errno.h>
#include <err.h>
#include <babolo/BLINflag.h>
#include "multilar.h"

#define LOCK(A)
#define UNLOCK(A)
#define SHIFT(A, B, C) mular_walk(mular, (A), (B), (C), stderr, NULL, mular_plus, NULL, NULL, NULL)

#define Sa 0x00000001U /* (segment->index >= indx)                                */
#define Ss 0x00000002U /* (segment->index < idx->index)     */
#define Se 0x00000004U /* (segment->index == idx->index)       */
#define Si 0x00000008U /* (segmin <= indx)                                        */
#define Sr 0x00000010U /* (res)                                      */
#define Sx 0x00000020U /* (segi < SIZE(level))                    */
#define So 0x00000040U /* (indx <= idx->index)                                    */
#define S_full   0x3FU

#define Cp 0x00800000U /*         */
#define Cr 0x00400000U /*   mular_get */
#define Cs 0x00200000U /*     */
#define Cd 0x00100000U /*  mular_dub           */
#define Cy 0x00080000U /*       */
#define Ci 0x00040000U /*           */
#define Ce 0x00020000U /*              */
#define C_state  0xFFU /*            */

enum states {d0, de, dx};
/* d0 -   
   de -  
   dx - 
 */

static void  mular_plus __P((mular_descriptor *, size_t, mular_indir *, int, int, FILE *));
static void *mular_dub  __P((mular_descriptor *, size_t, size_t, mular_indir *, int));

static const u_int32_t automa[S_full + 1][dx] =
/*        d0        de */
{ {   Cd| dx,   Cd| dx} /*      */
, {   Cd| dx,   Cd| dx} /*     a*/
, {Ce   | dx,Ce   | dx} /*    s */
, {Ce   | dx,Ce   | dx} /*    sa*/
, {Ce   | dx,Ce   | dx} /*   e  */
, {Ce   | dx,Ce   | dx} /*   e a*/
, {Ce   | dx,Ce   | dx} /*   es */
, {Ce   | dx,Ce   | dx} /*   esa*/
, {   Cd| dx,   Cd| dx} /*  i   */
, {   Cd| dx,   Cd| dx} /*  i  a*/
, {Ce   | dx,Ce   | dx} /*  i s */
, {Ce   | dx,Ce   | dx} /*  i sa*/
, {Ce   | dx,Ce   | dx} /*  ie  */
, {Ce   | dx,Ce   | dx} /*  ie a*/
, {Ce   | dx,Ce   | dx} /*  ies */
, {Ce   | dx,Ce   | dx} /*  iesa*/
, {   Cy| dx,   Cy| dx} /* r    */
, {   Cy| dx,   Cy| dx} /* r   a*/
, {Ce   | dx,Ce   | dx} /* r  s */
, {Ce   | dx,Ce   | dx} /* r  sa*/
, {Ce   | dx,Ce   | dx} /* r e  */
, {Ce   | dx,Ce   | dx} /* r e a*/
, {Ce   | dx,Ce   | dx} /* r es */
, {Ce   | dx,Ce   | dx} /* r esa*/
, {   Cy| dx,   Cy| dx} /* ri   */
, {   Cy| dx,   Cy| dx} /* ri  a*/
, {Ce   | dx,Ce   | dx} /* ri s */
, {Ce   | dx,Ce   | dx} /* ri sa*/
, {Ce   | dx,Ce   | dx} /* rie  */
, {Ce   | dx,Ce   | dx} /* rie a*/
, {Ce   | dx,Ce   | dx} /* ries */
, {Ce   | dx,Ce   | dx} /* riesa*/
, {Cp|Ci| de,   Cd| dx} /*x     */
, {   Ci| de,   Cd| dx} /*x    a*/
, {   Ci| d0,   Cd| dx} /*x   s */
, {   Ci| d0,   Cd| dx} /*x   sa*/
, {   Ci| de,   Cd| dx} /*x  e  */
, {   Ci| de,   Cd| dx} /*x  e a*/
, {Ce   | dx,Ce   | dx} /*x  es */
, {Ce   | dx,Ce   | dx} /*x  esa*/
, {Cp|Ci| de,   Cd| dx} /*x i   */
, {Cr|Ci| de,   Cd| dx} /*x i  a*/
, {   Ci| d0,   Cd| dx} /*x i s */
, {Cr|Ci| d0,   Cd| dx} /*x i sa*/
, {   Ci| de,   Cd| dx} /*x ie  */
, {Cr|Ci| de,   Cd| dx} /*x ie a*/
, {Ce   | dx,Ce   | dx} /*x ies */
, {Ce   | dx,Ce   | dx} /*x iesa*/
, {Cp|Ci| de,   Cy| dx} /*xr    */
, {Cs|Ci| de,   Cy| dx} /*xr   a*/
, {   Ci| d0,   Cy| dx} /*xr  s */
, {Cs|Ci| d0,   Cy| dx} /*xr  sa*/
, {   Ci| de,   Cy| dx} /*xr e  */
, {Cs|Ci| de,   Cy| dx} /*xr e a*/
, {Ce   | dx,Ce   | dx} /*xr es */
, {Ce   | dx,Ce   | dx} /*xr esa*/
, {Cp|Ci| de,   Cy| dx} /*xri   */
, {Cs|Ci| de,   Cy| dx} /*xri  a*/
, {   Ci| d0,   Cy| dx} /*xri s */
, {Cs|Ci| d0,   Cy| dx} /*xri sa*/
, {   Ci| de,   Cy| dx} /*xrie  */
, {Cs|Ci| de,   Cy| dx} /*xrie a*/
, {Ce   | dx,Ce   | dx} /*xries */
, {Ce   | dx,Ce   | dx} /*xriesa*/
};

#ifdef DEBUG
static mular_indir *ctrl[DEBUG] =
{ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0
, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0
};

static void *
mular_get(mular_descriptor *mular, size_t indx, size_t mindx, mular_indir *idx, int level) {
    void *res;

#   define blin_internal_flags ((mular ? mular->flags : mular_default) & BLIN_MASK)
    ifBLIN_QX3( "+ index=%"BLIN_U"(%"BLIN_U"..%"BLIN_U") level=%u size=%"BLIN_U
              , indx, mindx, idx->index, level, SIZE(level)
              );
    errno = 0;
    if  (indx < mindx || indx > idx->index || level < 1) {
        ifBLIN_QX0( " index=%"BLIN_U"(%"BLIN_U"..%"BLIN_U") level=%u size=%"BLIN_U
                  , indx, mindx, idx->index, level, SIZE(level)
                  );
        res = NULL;
        errno = EINVAL;
    } else if (level == 1) {
        if  (idx->index >= mindx + SIZE(level)) {
            ifBLIN_QX1(" level=%u size=%"BLIN_U" idx->index-mindx=%"BLIN_U"-%"BLIN_U
                      , level, SIZE(level), idx->index, mindx
                      );
            res = NULL;
        } else {
            res = ATERM(*idx, indx - mindx);
            if  (idx->index != indx) {
                TCP(ATERM(*idx, indx - mindx + 1), res, idx->index - indx);
                ifMULAR_ZERO TZR(res, 1);
            }
            idx->index++;
        }
    } else {
        size_t segi, segmin, nextmin = 0;
        mular_indir *segment = NULL;
        u_int32_t control, class, state;

        for ( segi = 0, segmin = mindx, res = NULL, control = 0, state = d0
            ; state < dx
            ; state = control & C_state
            ) {
            class = 0;
            if  (res) class |= Sr;
            if  (segmin <= indx) class |= Si;
            if  (segi < SIZE(level)) class |= Sx;
            control = automa[class][state];
            ifBLIN_QX3( ".%08X %08X i=%"BLIN_U"<%"BLIN_U"(%d) res=%"BLIN_X
                        " (min=%"BLIN_U")<=(indx=%"BLIN_U")"
                      , class, control, segi, SIZE(level), level, BLIN_I(res), segmin, indx
                      );
            if  (control & Cp) {
                segment = &NEXT(*idx)[segi];
                nextmin = segment->index;
                if  (segment->index == idx->index) class |= Se;
                if  (segment->index < idx->index) class |= Ss;
                if  (segment->index >= indx) class |= Sa;
                control = automa[class][state];
                ifBLIN_QX3( "@%08X %08X %"BLIN_U"<=%"BLIN_X"->%"BLIN_U"< =%"BLIN_U
                          , class, control, indx, BLIN_I(segment), segment->index, idx->index
                          );
            }
            if  (control & Cr) res = mular_get(mular, indx, segmin, segment, level - 1);
            if  (control & Cs) SHIFT(mindx, segment, level - 1);
            if  (control & Cd) res = mular_dub(mular, indx, mindx, idx, level);
            if  (control & Cy) idx->index++;
            if  (control & Ci) {
                segmin = nextmin;
                segi++;
            }
            if  (control & Ce) {
                ifBLIN_QX3( " Internal error class=%08X control=%08X"
                          , class, control
                          );
    }   }   }
    ifBLIN_QX3("- %"BLIN_X, BLIN_I(res));
    return(res);
#   undef blin_internal_flags
}

static void
mular_place(mular_indir *p, int n) {
    int i, j;
    for (j = 0; j < n; j++) {
        for (i = 0; i < DEBUG; i++) {
            if  (&p[j] == ctrl[i]) break;
            if  (!ctrl[i]) {
                ctrl[i] = &p[j];
                break;
}   }   }   }
#endif /* DEBUG */

static void
mular_plus(mular_descriptor *mular, size_t mindx, mular_indir *idx, int level, int ex, FILE *out) {
#   define blin_internal_flags ((mular ? mular->flags : mular_default) & BLIN_MASK)
    (void)mindx;
    (void)ex;
    (void)out;
    idx->index++;
    ifBLIN_QX2(" %"BLIN_X"(%zu) level=%u", BLIN_I(idx), idx->index, level);
#   undef blin_internal_flags
}

static size_t
mular_todiv(mular_descriptor *mular, size_t relidx, int level) {
    ssize_t todiv;
    ssize_t delta;

#   define blin_internal_flags ((mular ? mular->flags : mular_default) & BLIN_MASK)
    todiv = (ssize_t)((mular->flags & MULAR_HALF) ? SIZE(level) / 2 : relidx);
    if  (!(mular->flags & MULAR_STRI)) {
        if  (SIZE(level) / 2 < relidx) {
            delta = (ssize_t)(relidx - SIZE(level) / 2);
            delta = -delta / 4;
        } else {
            delta = (SIZE(level) / 2 - relidx) / 4;
        }
        todiv += (mular->flags & MULAR_HALF) ? -delta : delta;
    }
    return((size_t)todiv);
#   undef blin_internal_flags
}

static void *
mular_div(mular_descriptor *mular, size_t indx, size_t mindx, mular_indir dubl[2], int level) {
    size_t todiv, relidx;
    void *res;

#   define blin_internal_flags ((mular ? mular->flags : mular_default) & BLIN_MASK)
    ifBLIN_QX3( " + %"BLIN_U"(%"BLIN_U"..%"BLIN_U") [%"BLIN_X" %"BLIN_X"] level=%d"
              , indx, mindx, dubl[0].index, BLIN_I(&dubl[0]), BLIN_I(&dubl[1]), level
              );
    if  (indx < mindx || indx > dubl[0].index || level < 1) {
        ifBLIN_QX1( " index=%"BLIN_U"(%"BLIN_U"..%"BLIN_U") level=%u size=%"BLIN_U
                  , indx, mindx, dubl[0].index, level, SIZE(level)
                  );
        errno = EINVAL;
        res = NULL;
    } else if (!(TERM(dubl[1]) =  (mular->flags & MULAR_ZERO)
                                ? calloc(SIZE(level), level == 1 ? mular->unit : sizeof(mular_indir))
                                : malloc(level == 1 ? LTERM(SIZE(level)) : LNEXT(SIZE(level)))
              ) ) {
        ifBLIN_QW0("no mem");
        res = NULL;
    } else if (level == 1) {
        relidx = indx - mindx;
        todiv = mular_todiv(mular, relidx, level);
        ifBLIN_QO4 mular_displine( mular, relidx, mindx, dubl, level, stderr
                                 , MULAR_DISPLINE_D | MULAR_DISPLINE_P | (MULAR_DISPLINE_N & todiv)
                                 );
        if  (mindx + todiv > dubl[0].index) todiv = dubl[0].index - mindx;
        dubl[1].index = dubl[0].index + 1;

        /* todiv     ,  relidx,         *
         *          *
         *    relidx                 */
        if  (relidx < todiv) {
            ifBLIN_QX2("#1 idx=%"BLIN_U", div=%"BLIN_U, relidx, todiv);
            res = TD(0, relidx);
            TCP(TD(1, 0), TD(0, todiv), SIZE(level) - todiv);
            TCP(TD(0, relidx + 1), res, todiv - relidx);
            ifMULAR_ZERO {
                TZR(TD(0, todiv + 1), SIZE(level) - todiv - 1);
                TZR(res, 1);
            }
            dubl[0].index = mindx + todiv + 1;
        } else if (relidx > todiv) {
            ifBLIN_QX2("#2 idx=%"BLIN_U", div=%"BLIN_U, relidx, todiv);
            res = TD(1, relidx - todiv);
            TCP(TD(1, 0), TD(0, todiv), relidx - todiv);
            TCP(TD(1, relidx - todiv + 1), TD(0, relidx), SIZE(level) - relidx);
            ifMULAR_ZERO TZR(TD(0, todiv), SIZE(level) - todiv);
            dubl[0].index = mindx + todiv;
        } else if (todiv == 0) {                                                  /* relidx == todiv */
            ifBLIN_QX2("#3 idx=%"BLIN_U", div=%"BLIN_U, relidx, todiv);
            res = TD(0, 0);                                                 /* may be spap 0 and 1 ? */
            TCP(TD(1, 0), TD(0, 0), SIZE(level));
            ifMULAR_ZERO TZR(res, SIZE(level));
            dubl[0].index = mindx + 1;
        } else if (todiv == dubl[0].index) {                                      /* relidx == todiv */
            ifBLIN_QX2("#4 idx=%"BLIN_U", div=%"BLIN_U, relidx, todiv);
            res = TD(1, 0);
        } else {                                                                  /* relidx == todiv */
            ifBLIN_QX2("#5 idx=%"BLIN_U", div=%"BLIN_U, relidx, todiv);
            res = TD(1, 0);
            TCP(TD(1, 1), TD(0, todiv), SIZE(level) - todiv);
            ifMULAR_ZERO TZR(TD(0, todiv), SIZE(level) - todiv);
            dubl[0].index = mindx + todiv;
        }
        ifBLIN_QO4 mular_displine(mular, relidx, mindx, dubl, level, stderr, 2);
        ifBLIN_QX2( "0=%"BLIN_X"[%"BLIN_U"] 1=%"BLIN_X"[%"BLIN_U"] r=%"BLIN_X
                  , BLIN_I(TERM(dubl[0])), dubl[0].index
                  , BLIN_I(TERM(dubl[1])), dubl[1].index, BLIN_I(res)
                  );
    } else {
        ssize_t cont;
        mular_indir *t;
        size_t i, j;

        MULAR_INBUG(NEXT(dubl[1]), 1);
        for (i = 0, j = mindx; ((cont = ICONN(dubl, i, t)) >= 0); i++) {
            ifBLIN_QX4( " i=%"BLIN_U" j=%"BLIN_U" cont=%"BLIN_D
                        " t->index=%"BLIN_U" indx=%"BLIN_U" t=%"BLIN_X
                      , i, j, cont, t->index, indx, BLIN_I(t)
                      );
            if  (t->index > indx) break;
            if  (!cont) break;
            j = t->index;
        }
        relidx = i++;
        todiv = mular_todiv(mular, relidx, level);
        dubl[1].index = dubl[0].index + 1;
        ifBLIN_QO4 mular_displine( mular, relidx, mindx, dubl, level, stderr
                                 , MULAR_DISPLINE_D | MULAR_DISPLINE_P | (MULAR_DISPLINE_N & todiv)
                                 );
        for (; cont && ((cont = ICONN(dubl, i, t)) >= 0);) {
            ifBLIN_QX3(" i=%"BLIN_U" cont=%"BLIN_D" t=%"BLIN_X, i, cont, BLIN_I(t));
            SHIFT(mindx, t, level - 1);
            i++;
            if  (!cont) break;
        }
        /* todiv     ,  relidx,         *
         *          *
         *    relidx                 *
         *  , ,   relidx            *
         *                                       */
        if  (relidx < todiv) {
            ifBLIN_QX2("+1 reli=%"BLIN_U" div=%"BLIN_U" i=%"BLIN_U, relidx, todiv, i);
            t = ND(0, relidx);
            NCP(ND(1, 0), ND(0, todiv), i - todiv);
            NCP(ND(0, relidx + 2), ND(0, relidx + 1), todiv - relidx - 1);
            ifMULAR_ZERO {
                NZR(ND(0, todiv + 1), SIZE(level) - todiv - 1);
                NZR(ND(0, relidx + 1), 1);
            }
            res = mular_div(mular, indx, j, t, level - 1);
            dubl[0].index = ND(0, todiv)->index;
        } else if (relidx > todiv) {
            ifBLIN_QX2("+2 reli=%"BLIN_U" div=%"BLIN_U" i=%"BLIN_U, relidx, todiv, i);
            t = ND(1, relidx - todiv);
            NCP(ND(1, 0), ND(0, todiv), relidx - todiv + 1);
            NCP(ND(1, relidx - todiv + 2), ND(0, relidx + 1), i - relidx - 1);
            ifMULAR_ZERO NZR(ND(0, todiv), SIZE(level) - todiv);
            res = mular_div(mular, indx, j, t, level - 1);
            dubl[0].index = ND(0, todiv - 1)->index;
        } else if (todiv == 0) {                                                  /* relidx == todiv */
            ifBLIN_QX2("+3 reli=%"BLIN_U" div=%"BLIN_U" i=%"BLIN_U, relidx, todiv, i);
            t = ND(0, 0);
            dubl[0].index = ND(0, 0)->index + 1;
            NCP(ND(1, 0), ND(0, 1), i - 1);
            ifMULAR_ZERO NZR(ND(0, 1), SIZE(level) - 1);
            res = mular_div(mular, indx, j, t, level - 1);
        } else if (todiv == i - 1) {                                              /* relidx == todiv */
            ifBLIN_QX2("+4 reli=%"BLIN_U" div=%"BLIN_U" i=%"BLIN_U, relidx, todiv, i);
            t = ND(1, 0);
            dubl[0].index = ND(0, todiv - 1)->index;
            NCP(ND(1, 0), ND(0, todiv), 1);
            ifMULAR_ZERO NZR(ND(0, todiv), 1);
            res = mular_div(mular, indx, j, t, level - 1);
        } else if (mular->flags & MULAR_UPPE) {                                   /* relidx == todiv */
            ifBLIN_QX2("+5 reli=%"BLIN_U" div=%"BLIN_U" i=%"BLIN_U, relidx, todiv, i);
            t = ND(1, 0);
            dubl[0].index = ND(0, todiv - 1)->index;
            NCP(ND(1, 0), ND(0, todiv), 1);
            NCP(ND(1, 2), ND(0, todiv + 1), i - todiv - 1);
            ifMULAR_ZERO {
                NZR(ND(0, todiv), SIZE(level) - todiv);
                NZR(ND(1, 1), 1);
            }
            res = mular_div(mular, indx, j, t, level - 1);
        } else {
            ifBLIN_QX2("+6 reli=%"BLIN_U" div=%"BLIN_U" i=%"BLIN_U, relidx, todiv, i);
            t = ND(0, todiv);
            dubl[0].index = t->index + 1;
            NCP(ND(1, 0), ND(0, todiv + 1), i - todiv - 1);
            ifMULAR_ZERO NZR(ND(0, todiv + 1), SIZE(level) - todiv - 1);
            res = mular_div(mular, indx, j, t, level - 1);
        }
        ifBLIN_QO4 mular_displine(mular, relidx, mindx, dubl, level, stderr, 2);
    }
    ifBLIN_QX3("- %"BLIN_X, BLIN_I(res));
    return(res);
#   undef blin_internal_flags
}

static void *
mular_dub(mular_descriptor *mular, size_t indx, size_t mindx, mular_indir *idx, int level) {
    void *res;

#   define blin_internal_flags ((mular ? mular->flags : mular_default) & BLIN_MASK)
    ifBLIN_QX3( "+ flags=%08X i=%"BLIN_U"[%"BLIN_U"..%"BLIN_U"]@%"BLIN_X" level=%d"
              , mular->flags, indx, mindx, idx->index, BLIN_I(idx), level
              );
    if  (indx < mindx || indx > idx->index) {
        ifBLIN_QX1( " index=%"BLIN_U"(%"BLIN_U"..%"BLIN_U") level=%u size=%"BLIN_U
                  , indx, mindx, idx->index, level, SIZE(level)
                  );
        errno = 0;
        res = NULL;
    } else if (level > mular->level || level < 2) {
        ifBLIN_QX0("level too big: %u > %u", level, mular->level);
        errno = EINVAL;
        res = NULL;
    } else {
        ssize_t cont;
        mular_indir *t, *f;
        size_t i, j, relidx, midx;

        for (i = 0; i < SIZE(level) && ICONT(idx, i) > 0; i++) {}
        if  (++i < SIZE(level)) {
            for (i = 0, j = mindx; ((cont = ICONN(idx, i, t)) >= 0); i++) {
                if  (t->index > indx) break;
                if  (!cont) break;
                j = t->index;
            }
            relidx = i++;
            midx = j;
            ifBLIN_QO4 {
                mular_displine( mular, relidx, midx, idx, level, stderr, MULAR_DISPLINE_P | 1);
                mular_displine( mular, relidx, midx, t, level - 1, stderr, MULAR_DISPLINE_P | 1);
            }
            for (; cont && ((cont = ICONN(idx, i, f)) >= 0);) {
                SHIFT(j, f, level - 1);
                i++;
                if  (!cont) break;
                j = f->index;
            }
            NCP(&t[2], &t[1], i - relidx - 1);
            ifMULAR_ZERO NZR(&t[1], 1);
            res = mular_div(mular, indx, midx, t, level - 1);
            idx->index++;
        } else {
            ifBLIN_QX0( "No space in table: %"BLIN_U" <= %"BLIN_U" level=%d for indx=%"BLIN_U
                      , SIZE(level), i, level, indx
                      );
            errno = EINVAL;
            res = NULL;
    }   }
    ifBLIN_QX3( "- i=%"BLIN_U"[%"BLIN_U"..%"BLIN_U"] level=%d res=%"BLIN_X
              , indx, mindx, idx->index, level, BLIN_I(res)
              );
    return(res);
#   undef blin_internal_flags
}

void *
mular_insert(mular_descriptor *mular, size_t indx) {
    void *res;

#   define blin_internal_flags ((mular ? mular->flags : mular_default) & BLIN_MASK)
    ifBLIN_QX2( "+ flags=%08X i=%"BLIN_U" up=%"BLIN_U" level=%d"
              , mular->flags, indx, mular->top.index, mular->level
              );
    LOCK(mular);
    errno = 0;
    res = NULL;
    if  (mular->flags & MULAR_FIXD) {
        ifBLIN_QX0("mular fixed");
        errno = EINVAL;
    } else if  (indx > mular->top.index) {
        ifBLIN_QX0("void array, index=%"BLIN_U, indx);
        errno = EINVAL;
    } else if (mular->level == 0) {
        res = mular->flags & MULAR_ZERO ? calloc(SIZE(1), mular->unit) : malloc(LTERM(SIZE(1)));
        if  (!res) {
            ifBLIN_QW0("mular_insert #4");
        } else {
            TERM(mular->top) = res;
            mular->level = 1;
            mular->top.index = 1;
        }
    } else if (!(res = mular_get(mular, indx, 0, &mular->top, mular->level))) {
        size_t i;

        if  (mular->level == 1) {
            i = mular->top.index;
        } else {
            for (i = 0; i < SIZE(mular->level) && ICONT(&(mular->top), i) > 0; i++) {}
        }
        if  (++i >= SIZE(mular->level)) {
        ;   if  (mular->lemax > mular->level) {
                mular_indir *t;

        ;       if  (!(t =  mular->flags & MULAR_ZERO
                          ? calloc(SIZE(mular->level + 1), sizeof(mular_indir))
                          : malloc(LNEXT(SIZE(mular->level + 1)))
                    ) ) {
        ;           ifBLIN_QW0("No mem #3");
        ;       } else {
        ;           MULAR_INBUG(t, 1);
        ;           NCP(t, &mular->top, 1);
        ;           NEXT(mular->top) = t;
        ;           mular->level++;
        ;           res = mular_dub(mular, indx, 0, &(mular->top), mular->level);
        ;       }
        ;   } else if  (mular->level == 1) {
        ;       if  (!(TERM(mular->top) = realloc(TERM(mular->top), LTERM(SIZE(mular->level)) * 2))) {
        ;           ifBLIN_QW0("No mem #1");
        ;           res = NULL;
        ;       } else {
        ;           ifMULAR_ZERO TZR(ATERM(mular->top, SIZE(mular->level)), SIZE(mular->level));
        ;           SIZE(mular->level) *= 2;
        ;           res = mular_get(mular, indx, 0, &(mular->top), mular->level);
        ;       }
        ;   } else {
        ;       if  (!(NEXT(mular->top) = realloc(NEXT(mular->top), LNEXT(SIZE(mular->level)) * 2))) {
        ;           ifBLIN_QW0("No mem #2");
        ;       } else {
        ;           MULAR_INBUG(NEXT(mular->top), 2);
        ;           ifMULAR_ZERO NZR(&NEXT(mular->top)[SIZE(mular->level)], SIZE(mular->level));
        ;           SIZE(mular->level) *= 2;
                    res = mular_dub(mular, indx, 0, &(mular->top), mular->level);
        }   }   }
        if  (!res) {
            ifBLIN_QX0("No res");
            errno = EINVAL;
    }   }
    ifBLIN_QX2( "- up=%"BLIN_U" level=%d res=%"BLIN_X
              , mular->top.index, mular->level, BLIN_I(res)
              );
    UNLOCK(mular);
    return(res);
#   undef blin_internal_flags
}

void *
mular_add(mular_descriptor *mular) {
    return(mular_insert(mular, mular->top.index));
}
